//========================================================================================
// Mer.hh
// ---------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
// $Id: //CBT-depot/CompBioTools++/includes/Mer.hh#4 $
//----------------------------------------------------------------------------------------
//
//========================================================================================
//.......1.........2.........3.........4.........5.........6.........7.........8.........9
//

#ifndef _MER
#define _MER

#include "CompBioTools.hh"
#include "Sequence3Bit.hh"


//----------------------------------------------------------------------------------------
// * Mer
//----------------------------------------------------------------------------------------
// Class to represent an n-mer up to 32bp with a 2-bit/bp encoding. This class is primarialy
// used for hashing n-mers where keeping the total space to a minimum is important.  Since 
// systems we use have 32 bit ints, Mer uses two ints to represent a 32bp sequence.  For 
// For efficiency at I/O, the encoding is masked directly from the ASCII:
//
//      Hex         Dec            Binary   Encoding Bits
// a = 61 (hex)    97 (dec)      0110 0001   0110 0(00)1   a -> 0
// c = 63 (hex)    99 (dec)      0110 0011   0110 0(01)1   c -> 1
// g = 67 (hex)    103 (dec)     0110 0111   0110 0(11)1   g -> 3
// t = 74 (hex)    116 (dec)     0111 0100   0111 0(10)0   t -> 2

// A = 41 (hex)    65 (dec)      0100 0001   0100 0(00)1   A -> 0
// C = 43 (hex)    67 (dec)      0100 0011   0100 0(01)1   C -> 1
// G = 47 (hex)    71 (dec)      0100 0111   0100 0(11)1   G -> 3 
// T = 54 (hex)    84 (dec)      0101 0100   0101 0(10)0   T -> 2 

// So, bits 0,2 effectively code for a,c,t,g


// 3-bit encoding to include N/X.  

// When we use 3-bit bases, would like encoding for n/x to be a 
// superset of the acgtACGT encoding.   Bit 3 is set for n/x where it 
// is not set for acgtACGT, so bit 3 is diagnostic and bits 1,2,3 can 
// be used as the 3-bit encoding. 

// n = 6e (hex)    110 (dec)     0110 1110   0110 (111)0   n -> 7
// x = 78 (hex)    120 (dec)     0111 1000   0111 (100)0   x -> 4
//
// N = 4e (hex)    78 (dec)      0100 1110   0100 (111)0   N -> 7
// X = 58 (hex)    88 (dec)      0101 1000   0101 (100)0   X -> 4
//
//
// 
// 
class Mer{
public:

  static int mMerSize;
  
  unsigned int mUpper32;
  unsigned int mLower32;

  Mer(){};
  Mer(Sequence3Bit &Seq,int merStart,bool &ContainsNX);
  Mer(Sequence &Seq);
  Mer(Sequence &Seq,int merStart);
  Mer(Sequence3Bit &Seq,int merStart);
  Mer(string s);
  
  // Methods for {A,C,G,T} = bit1,bit2 of ASCII
  void ReverseComplement(Mer &fwdMer);
  void ReverseComplement();
  inline unsigned int ComplementBits(unsigned int bits);

  inline unsigned int Base2Bit(char b);
  void ToSequence(Sequence &S);
  inline char Bit2Base(unsigned int x);


  bool operator==(const Mer &m) const {
	 return((m.mUpper32 == mUpper32) && (m.mLower32 == mLower32));
  }


  bool InSubset(unsigned int Subset);
  
  // DEPRECATED ==================
  // Methods  for {A,C,G,T} = {0,1,2,3}
  void ReverseComplement2(Mer &fwdMer);
  void ReverseComplement2();
  inline unsigned int Base2Bit2(char b);
  void ToSequence2(Sequence &S);
  inline char Bit2Base2(unsigned int x);
  //===============================
};


struct eqMer{
  bool operator()(const Mer a,const Mer b) const {
	 // The mers match in the same sense. 
	 if ((a.mUpper32 == b.mUpper32) && (a.mLower32 == b.mLower32)){
		return(true);
	 }else return(false);
  }
};

// An equal function that does not distinguish between fwd and 
// revcomp versions of a Mer. 
// 
struct eqMerSymmetric{

  bool operator()(const Mer a,const Mer b) const {
	 // The mers match in the same sense. 
	 if ((a.mUpper32 == b.mUpper32) && (a.mLower32 == b.mLower32)){
		return(true);
	 }else{
		// see if the reverse complement matches. 
		Mer rcb = b;
		rcb.ReverseComplement();
		if ((a.mUpper32 == rcb.mUpper32) && (a.mLower32 == rcb.mLower32)){
		  return(true);
		}else{
		  return(false);
		}
	 }
  }
};


#endif
