#include "CompBioTools.hh"
#include "ContigEvalUtils.hh"
#include "ContigFixupUtils.hh"

extern int trimWindow;
extern float trimHeight;

string AceIn = "";
string QualIn = "";
string AceOut="";
string TrimmedReadsOut="";
string GraphFile = "";
string ProblemFile = "";

bool gBugFlag = false;


int  OptMinMiddle = 50000;

class SplitPoint{
public:
  int mLeft;
  int mRight;

  friend bool operator<(const SplitPoint a,const SplitPoint b){
	 return(a.mLeft<b.mLeft);
  }
};


void PrintHelp()
{
  printf("
      atlas-splitbadcontigs  -a AceIn  -q ReadQualityIn -o AceOut {options} 

      A moving average of the template coverage is computed starting at the ends
      and going inward.  The window size for this moving average is given by trimWindow, 
      and is expressed in base pairs.   The trim cutoff is specified as the first place
      where the moving average of the template coverage is greater than trimHeight.
      The contig is split into two contigs around this 'template crash'.  Reads which 
      overlap the left and right boundary of the crash region go to the left or right
      if they are anchored there by a matched mate, but otherwise are simply dropped. 

      When a ProblemFile is given with the -p option, a totally different set of more
      subtle split rules are used that aren't just based on template crashes. 

      -a AceFile
      -q Read qualities (for new consensus computation). 
      -o Ace output file name. 

      ----------------

      -r TrimmedReadListOut 
      -w trimWindow  (for default template crash mode, i.e. not used with -p)
      -c trimCoverage (for default templat crash mode, i.e. not used with -p)

      -g graph output which includes Rep_node: specifiers.  When given 
         only nodes listed in the Rep_node: list will be trimmed, others will 
         be left alone.  This option excludes -p.  With this option, splits are
         on template crashes only.   


      -p Problem file.  This is a file that identifies regions of suspicious patterns
         of external links.  These are probably problems, so if at all possible, a 
         split will be done on each of these problems.  This option excludes -g.
         With this option, a series of rules are applied to determine if the suggested 
         problem area is real and, if so, how to split. 


      -M Minimum size of zero template coverage contig to keep after splitting. 


      Version History
      ------------------

      Version 1.36  Added synthetic BS lines to newly created contigs so that 
                    the output will be fully compatible with Consed.  

      Version 1.34  Fixed bug in AdjustSplitPoints due to uninitialized variable. 

      Version 1.32  When a contig was split and the left piece has zero size, it would
                    be deleted.  This, however, changes the index of the contigs in the 
                    Ace file. The program looped through contig indexes and so would sometimes
                    skip a contig that should be split.  Now I loop through by contig name, 
                    so that it is insensitive to changes in the contig index. 

      Version 1.30  Fixed merged-pairs code. 

      Version 1.24  Drop any contig from a split when zero template coverage and 
                    less than -M length. 

      Version 1.22  Add option to keep middle only when large.  

      Version 1.20  Add support to keep reads in middle of a crash.  Reorganize code a bit. 

      Version 1.18  Added support for a checkpoint (problem) file, that suggests possible 
                    problem regions.  This is to allow more subtle splitting than just on 
                    template crashes alone. 
      Version 1.17  Reorganized code. 
      Version 1.16  Replaced memory hogging add contig procedure with simple resize()
                    (Note: previously this didn't work, but the bug fixed in 1.14 also 
                     made it possible to go back to this method).  
      Version 1.14  Fixed bug where leftmostAlteredConsensus was sometimes negative
                    causing memory corruption.  Software is much more stable now. 
      Version 1.12  Fixed bug reading quality values where duplicate reads in 
                    quality input resulted in simply concatenating the two qualities
                    together.   
      Version 1.11

\n");
	 exit(1);
}



//---------------------------------------------------
// * SetupOptions
//---------------------------------------------------
//
void SetupOptions(int argc, char**argv)
{
  // Print it for the record
  for(int i = 0;i< argc;i++) cout<<argv[i]<<" ";
  cout<<endl;

  // Create parser object and define valid options
  // Options with a : require a parameter. 
  GetOption Parser(argc,argv,"a:q:o:r:w:c:g:p:M:h");

  // Handle the options...
  char theOption;
  while((theOption = Parser.NextOption()) != OPTION_EOF){
    switch(theOption){
	 case 'a': AceIn = Parser.GetStringArg();break;
	 case 'q': QualIn = Parser.GetStringArg();break;
	 case 'o': AceOut = Parser.GetStringArg();break;
	 case 'r': TrimmedReadsOut = Parser.GetStringArg();break;	
	 case 'w': trimWindow = Parser.GetIntArg();break;
	 case 'c': trimHeight= Parser.GetFloatArg();break;
	 case 'g': GraphFile = Parser.GetStringArg();break;
	 case 'p': ProblemFile = Parser.GetStringArg();break;
	 case 'M': OptMinMiddle = Parser.GetIntArg();break;
    default: PrintHelp(); exit(1); break;
    }
  }

  if ((GraphFile != "") && (ProblemFile != "")){
	 cerr<<" Can't specify both a graph file and a problem file. "<<endl;
	 PrintHelp();
	 exit(1);
  }

}

void SortSplits(vector<int> &lefts, vector<int> &rights)
{
  if (lefts.size() == 0){
	 if (rights.size() != 0){
		cerr<<"SortSplits:: left and right sizes don't match."<<endl;
	 }
	 return;
  }


  vector<SplitPoint> Splits;
  for(int i = 0;i< lefts.size();i++){
	 SplitPoint temp;
	 temp.mLeft = lefts[i];
	 temp.mRight = rights[i];
	 Splits.push_back(temp);
  }

  sort(Splits.begin(),Splits.end());

  for(int i = 0;i< Splits.size();i++){
	 lefts[i] = Splits[i].mLeft;
	 rights[i] = Splits[i].mRight;
  }
}


//--------------------------------------
// * main
//--------------------------------------
// 
int main(int argc,char** argv)
{
  SetupOptions(argc,argv);
  
  vector<string> RemovedReads;
  
  StringSet GraphContigs,ProblemContigs;
  if (GraphFile != ""){
	 cout<<"Reading rep contig names..."<<flush;
	 GetProblemContigs(GraphFile,GraphContigs);
	 cout<<GraphContigs.size()<<"done."<<endl;
  }
  
  ContigChecks Checks;
  if (ProblemFile != ""){
	 cout<<"Reading problem contig file..."<<flush;
	 Checks.Read(ProblemFile);
	 for(int i = 0;i< Checks.size();i++) ProblemContigs.Add(Checks[i].mContigName);	 
	 cout<<"done."<<endl;
  }
  
  // If there is no problem file or graph file, then we try to split all 
  // contigs.  
  bool bDoAllContigs;
  bDoAllContigs = false;

  if ((ProblemFile == "") && (GraphFile == "")) bDoAllContigs = true;
  
  cout<<"Reading "<<AceIn<<"..."<<flush;
  AceFile Ace;
  
  Ace.ReadAll(AceIn,QualIn);
  cout<<". done."<<endl;

  StringSet RemoveContigList;
  
	// Good pairs of coordinates are a close left/right pair. 
	// Here we define close as within 5Kbp.  Some rules only 
	// apply if the merged pair is even closer. 
	Checks.MergePairs(5000);

  vector<string> RemoveReads;

	// Create a list of the contig names to look up. 
	// MultiSplitContig will delete contigs and so we can't count 
	// on the index within the Ace file staying the same. 
	vector<string> ContigNames;
	for(int i = 0;i< Ace.size();i++) ContigNames.push_back(Ace[i].mName);

  for(int c = 0;c < ContigNames.size();c++){		
		int cIdx = Ace.CName2CIdx(ContigNames[c]);

		if ((cIdx < 0) || (cIdx >= Ace.size())){
			cerr<<"ERROR: ContigIndex out of bounds:"<<cIdx<<" Ace.size="<<Ace.size()<<"\t"<<ContigNames[c]<<endl;
			continue;
		}
		
		Contig &C = Ace[cIdx];
	 
	 // Only examine the contigs specified....
	 if (bDoAllContigs || GraphContigs.Contains(C.mName) || ProblemContigs.Contains(C.mName)){
		
		if (ProblemContigs.Contains(C.mName)){		 

		  vector<int> leftSplits,rightSplits;
		  IdentifySplits(C,Checks,leftSplits,rightSplits);

		  SortSplits(leftSplits,rightSplits);
		  
		  Ace.MultiSplitContig(OptMinMiddle,cIdx,leftSplits,rightSplits);

		}else{

		  vector<int> leftSplits,rightSplits;
		  
			cout<<"*** Temporarialy disabled splitting without checkpoint."<<endl;
		  
		  // Generic split crashes that looks for a template crash 
		  // in a contig and splits the contig there.  The split contigs
		  // are also examined for crashes, so a contig may be split several times. 
		  //SplitCrashes(Ace,c,RemovedReads,RemoveContigList);
		}
	 }else{
		//cout<<"Skipping contig "<<C.mName<<endl;
	 }
  }
  cout<<"done."<<endl;

  cout<<"Writing "<<AceOut<<"..."<<flush;
  Ace.Write(AceOut,RemoveContigList);
  cout<<"done."<<endl;

  cout<<"Writing "<<TrimmedReadsOut<<"..."<<flush;
  WriteFOF(TrimmedReadsOut,RemovedReads);
  cout<<"done."<<endl;

  exit(0);

}













