#include "CompBioTools.hh"
#include "ContigEvalUtils.hh"

string AceIn = "";
string AceOut="";
string TrimmedReadsOut="";
string GraphFile = "";
string ReadQualIn = "";

int MinKeepSize = 2000;

extern int trimWindow;
extern float trimHeight;



void PrintHelp()
{
  printf("
      atlas-trimphraptails  -a AceIn -q ReadQualities  -o AceOut {options} 

      -a AceFile
      -q Read qualities (for new consensus computation). 
      -o Ace output file name. 

      ----------------

      -r TrimmedReadListOut 
      -w trimWindow 
      -c trimCoverage
      -g graph output which includes Rep_node: specifiers.  When given 
         only nodes listed in the Rep_node: list will be trimmed, others will 
         be left alone.   Unlike previous versions, starting with version 1.1 
         'node' is taken to be just one side of a contig.  So if Contig63_right 
         is the only rep node, only the right half of Contig63 will be trimmed. 

      -M Minimum size of trimmed contig to keep.  If either (or both) of 
         contig and trim portion of contig are smaller than this, they are dropped. 

      A moving average of the template coverage is computed starting at the ends
      and going inward.  The window size for this moving average is given by trimWindow, 
      and is expressed in base pairs.   The trim cutoff is specified as the first place
      where the moving average of the template coverage is greater than trimHeight. 


      Version 1.16
      This version should produce Consed readable files.  It includes synthetic BS 
      line generation. 

      \n");
	 exit(1);
}



//---------------------------------------------------
// * SetupOptions
//---------------------------------------------------
//
void SetupOptions(int argc, char**argv)
{
  // Print it for the record
  for(int i = 0;i< argc;i++) cout<<argv[i]<<" ";
  cout<<endl;

  // Create parser object and define valid options
  // Options with a : require a parameter. 
  GetOption Parser(argc,argv,"q:a:o:r:w:c:g:M:h");

  // Handle the options...
  char theOption;
  while((theOption = Parser.NextOption()) != OPTION_EOF){
    switch(theOption){
	 case 'a': AceIn = Parser.GetStringArg();break;
	 case 'o': AceOut = Parser.GetStringArg();break;
	 case 'r': TrimmedReadsOut = Parser.GetStringArg();break;	
	 case 'q': ReadQualIn = Parser.GetStringArg();break;
	 case 'w': trimWindow = Parser.GetIntArg();break;
	 case 'c': trimHeight= Parser.GetFloatArg();break;
	 case 'g': GraphFile = Parser.GetStringArg();break;
	 case 'M': MinKeepSize = Parser.GetIntArg();break;
    default: PrintHelp(); exit(1); break;
    }
  }
}

//--------------------------------------
// * GetProblemContigs
//--------------------------------------
// 
void GetProblemContigs(string GraphFile,String2StringMap &ProblemContigs)
{
  ifstream fin;
  fin.open(GraphFile.c_str());
  if (fin.fail()){
	 cerr<<"Unable to open "<<GraphFile<<endl;
	 exit(1);
  }

  do{
	 string line;
	 GetLine(fin,line);
	 if (Contains(line,"Rep_node:")){
		int idx0 = line.find(":");
		string line2 = line.substr(idx0+1,line.size()-idx0-1);
		
		int idx1 = line2.find("_");
		int idx2 = line2.find("_",idx1+1);

		// If there is only one underscore. KJD Why would there be two? 
		if (idx2 < 0){
		  string contigName = line2.substr(0,idx1);
		  StripWhiteSpace(contigName);
		  cout<<"Adding "<<contigName<<" to problem contigs list."<<endl;
		  
		  if (ProblemContigs.Contains(contigName)){
			 string temp =  ProblemContigs[contigName];
			 if (Contains(line,"left")) temp+="left";
			 if (Contains(line,"right")) temp+="right";
			 ProblemContigs[contigName] = temp;
		  }else{
			 string temp;
			 if (Contains(line,"left")) temp+="left";
			 if (Contains(line,"right")) temp+="right";
			 ProblemContigs[contigName] = temp;
		  }			 			 			 
		}
	 }
  }while(!fin.eof());
  
}

/*
void RemoveContigs(AceFile &A,vector<string> &RList)
{
  cout<<"Remove contigs Beforesize="<<A.size()<<"\t"<<endl;
  for(int i = 0;i< RList.size();i++){

	 // Search for the contig anew each time since this will
	 // affect a renumbering. 
	 for(int c = 0;c< A.size();c++){
		if (A[c].mName == RList[i]){
		  cout<<"size="<<A.size()<<"\tRemove "<<A[c].mName<<flush;
		  A.RemoveContig(c); // delete the entire contig.
 		  cout<<"\tsize="<<A.size()<<endl;
		}
	 }
  }
  cout<<"Remove contigs Aftersize="<<A.size()<<"\t"<<endl;
  
}
*/

// KJD Hack to handle the split contigs
// This function adds the names of the split contigs to the 
// ProblemContigs list. 
void AddSplitContigs(AceFile &Ace,String2StringMap &ProblemContigs)
{
  vector<string> SplitContigs;
  String2StringMap::iterator it;
  for(int c = 0;c< Ace.size();c++){
	 string cName = Ace[c].mName;
	 for(it = ProblemContigs.begin();it != ProblemContigs.end();it++){
		string OriginalCName = (*it).first;
		if (Contains(cName,OriginalCName)){
		  SplitContigs.push_back(cName);
		}
	 }
  }

  for(int i = 0;i< SplitContigs.size();i++){
	 ProblemContigs.Add(SplitContigs[i],"leftright");
  }
}


//--------------------------------------
// * main
//--------------------------------------
// 
int main(int argc,char** argv)
{
  SetupOptions(argc,argv);

  AceFile Ace;
  vector<string> RemovedReads;

  String2StringMap ProblemContigs;
  if (GraphFile != ""){
	 cout<<"Reading problem contig names..."<<flush;
	 GetProblemContigs(GraphFile,ProblemContigs);
	 cout<<ProblemContigs.size()<<"done."<<endl;
  }

  String2StringMap::iterator it;
  for(it = ProblemContigs.begin();it != ProblemContigs.end();it++){
	 cout<<"ProblemContigs\t"<<(*it).first<<"\t"<<(*it).second<<endl;
  }

  // A hack to add any split contigs to the problem contig list. 
  //AddSplitContigs(Ace,ProblemContigs);

  cout<<"Reading "<<AceIn<<"..."<<flush;
  //  Ace.Read(AceIn);
  Ace.ReadAll(AceIn,ReadQualIn);
  cout<<". done."<<endl;

  StringSet RemoveContigList;
  cout<<"Trimming contigs "<<endl;

  vector<string> DeleteList;

  for(int c = 0;c < Ace.size();c++){
	 Contig &C = Ace[c];
	 if ((ProblemContigs.Contains(C.mName)) || (GraphFile == "")){
		vector<int> TCover;
		vector<int> RCover;
		
		TCover.clear();
		RCover.clear();
		C.GetTemplateCoverage(TCover);
		C.GetCoverage(RCover);
		
		// Find low template coverage tails		
		int leftTail=-1,rightTail=-1;

		string sides = ProblemContigs[C.mName];
		if (Contains(sides,"left")) FindLeftTail(TCover,leftTail);
		else leftTail = 0;

		if (Contains(sides,"right")) FindRightTail(TCover,rightTail);
		else rightTail = TCover.size();
												 		
		cout<<C.mName<<"\tsides="<<sides<<"\t"<<C.size()<<"\tleftTail: "<<leftTail;
		cout<<"\trightTail: "<<rightTail<<"\t";

		bool bDegenerate = ((leftTail <= 0) && (rightTail <=0));
		bool bValid = ((leftTail>0) && (rightTail > 0) && (rightTail < C.size()));
		int  tsize = rightTail - leftTail;
		
		//cout<<"C.size="<<C.size()<<endl;
		//cout<<"bDegenerate="<<bDegenerate<<"\tbValid="<<bValid<<"\ttsize="<<tsize<<endl;

		if ((bValid) && (tsize < 1500) || bDegenerate){
		  cout<<"Mark for removal: "<<C.mName<<"\t size="<<(rightTail-leftTail)<<endl;
		  RemoveContigList.Add(C.mName);
		}else{
		  Contig NewC1,NewC2;
		  C.SplitContig(leftTail,rightTail,NewC1,NewC2);
		  cout<<"C.size="<<C.size()<<" NewC1.size="<<NewC1.size()<<" NewC2.size="<<NewC2.size()<<endl;
		  
		  if (C.size() < MinKeepSize) DeleteList.push_back(C.mName);
		  if (NewC1.size() >= MinKeepSize){
			 Ace.resize(Ace.size()+1);
			 int newCIdx = Ace.size() -1;
			 Ace[newCIdx] = NewC1;
		  }

		  if (NewC2.size() >= MinKeepSize){
			 Ace.resize(Ace.size()+1);
			 int newCIdx = Ace.size() -1;
			 Ace[newCIdx] = NewC2;
		  }
		  
		  //C.RemoveTails(leftTail,rightTail,RemovedReads);
		}
	 }else{
		cout<<"Skipping contig "<<C.mName<<endl;
	 }
  }
  cout<<"done."<<endl;

  for(int i = 0;i< DeleteList.size();i++){
	 int deleteIdx = Ace.CName2CIdx(DeleteList[i]);
	 Ace.erase(Ace.begin()+deleteIdx);
	 Ace.CleanInitFromContigs(); 
  }

  //RemoveContigs(Ace,RemoveContigList);

  cout<<"Writing "<<AceOut<<"..."<<flush;
  Ace.Write(AceOut,RemoveContigList);
  cout<<"done."<<endl;

  // cout<<"Writing "<<TrimmedReadsOut<<"..."<<flush;
  //WriteFOF(TrimmedReadsOut,RemovedReads);
  //cout<<"done."<<endl;
}





// Some debug code. 

  /*
  for(int c = 0;c < Ace.size();c++){
	 for(int r = 0;r< Ace[c].mReads.size();r++){
		if (Ace[c].mReads[r].size() != Ace[c].mReadsQuality[r].size()){
		  cout<<"SIZE MISMATCH "<<Ace[c].mReads[r].mName<<endl;
		}
		for(int i = 0;i<Ace[c].mReads[r].size();i++){
		  if (Ace[c].mReadsQuality[r][i] > 200){
			 cout<<"ERROR "<<Ace[c].mReads[r].mName<<":";
			 cout<<Ace[c].mReadsQuality[r].mName<<"\t";
			 cout<<Ace[c].mReadSense[r]<<"\t"<<i<<"\t";
			 cout<<Ace[c].mReads[r].size()<<"\t"<<Ace[c].mReadsQuality[r].size()<<"\t";
			 cout<<Ace[c].mReads[r][i]<<"\t"<<Ace[c].mReadsQuality[r][i]<<"\t";
			 cout<<Ace[c].mReadsQuality[r][i+1]<<"\t"<<Ace[c].mReadsQuality[r][i+2]<<"\t";
			 cout<<Ace[c].mReadsQuality[r][i+3]<<endl;
		  }
		}
	 }
  }
  */

