#include "CompBioTools.hh"
#include "HGSCOracle.hh"

// An HGSC specific class for dealing with scaffolds. 
// Class can be found under HGSC++. 
#include "Scaffold.hh"

string OutRoot = "/users/pipeline/p-atlas/atlas-temp/fz02.1-super-linear/output";
string BactigRoot = "/users/pipeline/p-atlas/atlas-temp/freeze02-wavestitch/output";

int OptMinN = 50;
bool bUseFinal = true;

int MinScaffoldSize = 1;

// Means no problems were encountered.  
bool bSuccess = true;


// One global oracle connection. 
HGSCOracleConnect DB("ruichen/ruichen@gsc");

//----------------------------------------------------------------------------
// * BactigRecord
//----------------------------------------------------------------------------
//
class BactigRecord{
public:
	string mBactig;
	string mScaffold;
	string mContig;
	int    mStart;
	int    mEnd;
	char   mOrient;
	char   mOrdered;

	// Sort on multiple keys, in order:
	// Ordered
	// BactigID
	// ScaffoldID
	// StartLocation
	//
	friend const bool operator<(const BactigRecord &a,const BactigRecord &b){
		if (a.mOrdered != b.mOrdered){
			if (a.mOrdered == '+') return(true);
			else return(false);
		}

		// For unordered, sort by bactig+scaffold first before sorting by start location.
		if ((a.mOrdered == '-') && (b.mOrdered == '-')){
			string ScaffoldIDA = a.mBactig+a.mScaffold;
			string ScaffoldIDB = b.mBactig+b.mScaffold;		

			if (ScaffoldIDA != ScaffoldIDB){
				return(ScaffoldIDA<ScaffoldIDB);
			}
		}

		return(a.mStart < b.mStart);	 
	}

	void Print(ostream &fout){
		fout<<mOrdered<<"\t"<<mBactig<<"\t"<<mScaffold<<"\t"<<mContig<<"\t";
		fout<<mStart<<"\t"<<mEnd<<"\t"<<mOrient<<endl;
	}
};

//----------------------------------------------------------------------------
// * BactigRecord
//----------------------------------------------------------------------------
//
class BactigRecords : public vector<BactigRecord>{
public:
	string mName;

	void GetFromDB(HGSCOracleConnect &DBConnection,string SuperBactigName){

		mName = SuperBactigName;

		string superID,bactigID,scaffold,contig;
		int    startpos,endpos;
		char   orient,ordered;
		
		// Get all of the bactig IDs for this superbactig...
		string query; 
		query= "select * from superbactig_scaffold where superbactig='"+SuperBactigName+"'";
		cout<<"bactig record query="<<query<<endl;
		HGSCOracleStream q(DBConnection,query);

		while(!q.eof()){
			q>>superID>>bactigID>>scaffold>>contig>>startpos>>endpos>>orient>>ordered;

			BactigRecord R;
			R.mBactig = bactigID;
			R.mScaffold = scaffold;
			R.mContig = contig;
			R.mStart = startpos;
			R.mEnd = endpos;
			R.mOrient = orient;
			R.mOrdered = ordered;
			
			push_back(R);
		}

		sort(begin(),end());
	}

	void Print(ostream &fout){
		for(int i = 0;i< size();i++){
			(*this)[i].Print(fout);
		}
	}

};

//-------------------------------------------
// * GetContigSequence
//-------------------------------------------
// Gets the trimmed and depadded sequence for the named
// contig. 
// 
void GetContigSequence(AceFile &Ace,string CName,Sequence &Seq,SequenceQuality &Qual)
{
	
	Seq.clear();
	Qual.clear();
  for(int i = 0;i<Ace.size();i++){
	 if (Ace[i].mName == CName){
		Seq.mName = Ace[i].mName;
		for(int j = Ace[i].mTrimLeft;j <=Ace[i].mTrimRight;j++){
		  if (Ace[i][j] != '*'){
			 Seq.push_back(Ace[i][j]);
			 Qual.push_back(Ace[i].mConsensusQuality[j]);
		  }
		}
		cout<<"GotContigSequence "<<CName<<endl;
		return;
	 }
  }

  cout<<"ERROR:  Didn't find "<<CName<<endl;
  //for(int i = 0;i< Ace.size();i++){
	// cout<<"\t"<<Ace[i].mName<<endl;
	// }
}

//-------------------------------------------
// * AddN
//-------------------------------------------
//
void AddN(Sequence &Seq,int NumN){ 
  for(int i = 0;i< NumN;i++){
	 Seq.push_back('N');
  }
}

//-------------------------------------------
// * AddZeros
//-------------------------------------------
//
void AddZeros(SequenceQuality &SeqQual,int NumN){ 
  for(int i = 0;i< NumN;i++){
	 SeqQual.push_back(0);
  }
}


//---------------------------------------------------
// * ComputeGapFactor
//---------------------------------------------------
//
//
double ComputeGapFactor(string name,string description,
																	 BactigRecords &R)
{ 
	Sequence ScaffoldSeq;
	SequenceQuality ScaffoldQual;

	Scaffold SActual;

	long totalReportedGapSize = 0;
	long totalActualGapSize = 0;

	string currentBactigID = "";
	AceFile *pA = NULL;
	int aStart = 0;
	for(int i = 0;i< R.size();i++){
		BactigRecord &BR = R[i];
		
		// If the record is for a different bactig, read it in...
		if (currentBactigID != BR.mBactig){
			currentBactigID = BR.mBactig;
			
			if (pA == NULL){ 
				pA = new AceFile();
			}else {
				delete pA;
				pA = new AceFile();
			}
			
			// Read the ace file for this bactig...
			string AceName;
			AceName = BactigRoot+"/"+BR.mBactig+"/bactig.final.ace.gz";
			cout<<"Reading "+AceName<<"..."<<flush;
			pA->mContigTrim = 1;
			pA->Read(AceName);
			cout<<"done."<<endl;
		}
		
		Sequence TempSeq;
		SequenceQuality TempQual;
		GetContigSequence((*pA),BR.mContig,TempSeq,TempQual);
		
		if (TempSeq.size() == 0){   
			cerr<<"FAILED to find contig sequence"<<endl;
			bSuccess = false;
		}

		// Compute the true contig boundaries for a true scaffold graph. 
		ScaffoldContig CTemp;
		CTemp.mStart = CTemp.mEnd = 0;
		CTemp.mName = BR.mBactig+"_"+BR.mContig;
		CTemp.mSense = BR.mOrient;
		CTemp.mStart = aStart;
		CTemp.mEnd = aStart+TempSeq.size()-1;
		SActual.push_back(CTemp);
		
		aStart = (int) CTemp.mEnd+1;
				
		// Add N's when it's not the last contig in the scaffold...
		if (i < R.size()-1){			 
			int gapSize = (int)( R[i+1].mStart - R[i].mEnd);

			totalReportedGapSize +=gapSize;

			int NumN = max(gapSize,OptMinN);

			totalActualGapSize += NumN;
			aStart=aStart+NumN;			 
		}				
	}
	if (pA != NULL) delete pA;

	// Now compare the size of the scaffold that we've created
	// to the reported scaffold size.
	double createdSize = (double)SActual[SActual.size()-1].mEnd;
	double reportedSize = (double) R[R.size()-1].mEnd;
	double diff = createdSize - reportedSize;

	// I want the actual gap size to be shurnk/expanded by diff.  
	double desiredGapSize = totalActualGapSize - diff;

	double factor;
	if (totalActualGapSize != 0){
		factor = fabs((double)desiredGapSize/(double)totalActualGapSize);
	}else{
		factor = 1.0;
	}

	cout<<"createdSize="<<createdSize<<" reportedSize="<<reportedSize<<endl;
	cout<<"totalActualGapSize="<<totalActualGapSize<<"\ttotalReportedGapSize = "<<totalReportedGapSize<<endl<<" diff="<<diff<<" desirecGapSize="<<desiredGapSize<<endl;
	cout<<"factor=\t"<<factor<<endl;
	return(factor);
}


//---------------------------------------------------
// * GetSequenceFromBactigRecords
//---------------------------------------------------
//
//
void GetSequenceFromBactigRecords(string name,string description,
																	BactigRecords &R,SequenceVector &SeqsOut,
																	SequenceQualityVector &QualsOut,
																	Scaffold &SActual,double factor)
{ 
	Sequence ScaffoldSeq;
	SequenceQuality ScaffoldQual;

	ScaffoldSeq.mName = name;
	ScaffoldQual.mName = name;
	ScaffoldSeq.mDescription = description;
	ScaffoldQual.mDescription = description;

	SActual.mID = ScaffoldSeq.mName;

	string currentBactigID = "";
	AceFile *pA = NULL;
	int aStart = 0;
	for(int i = 0;i< R.size();i++){
		BactigRecord &BR = R[i];
		
		// If the record is for a different bactig, read it in...
		if (currentBactigID != BR.mBactig){
			currentBactigID = BR.mBactig;
			
			if (pA == NULL){ 
				pA = new AceFile();
			}else {
				delete pA;
				pA = new AceFile();
			}
			
			// Read the ace file for this bactig...
			string AceName;
			AceName = BactigRoot+"/"+BR.mBactig+"/bactig.final.ace.gz";
			cout<<"Reading "+AceName<<"..."<<flush;
			pA->mContigTrim = 1;
			pA->Read(AceName);
			cout<<"done."<<endl;
		}
		
		Sequence TempSeq;
		SequenceQuality TempQual;
		GetContigSequence((*pA),BR.mContig,TempSeq,TempQual);
		
		if (TempSeq.size() == 0){   
			cerr<<"FAILED to find contig sequence"<<endl;
			bSuccess = false;
		}else{
			if (BR.mOrient == '-'){
				TempSeq.ReverseComplement();
				TempQual.Reverse();
			}
		}    

		// Compute the true contig boundaries for a true scaffold graph. 
		ScaffoldContig CTemp;
		CTemp.mStart = CTemp.mEnd = 0;
		CTemp.mName = BR.mBactig+"_"+BR.mContig;
		CTemp.mSense = BR.mOrient;
		CTemp.mStart = aStart;
		CTemp.mEnd = aStart+TempSeq.size()-1;
		SActual.push_back(CTemp);
		
		aStart = (int) CTemp.mEnd+1;
		
		cout<<"Adding "<<TempSeq.size()<<" bases. "<<flush;
		ScaffoldSeq.Concatenate(TempSeq);
		cout<<" totalsize="<<ScaffoldSeq.size()<<endl;
		ScaffoldQual.Concatenate(TempQual);
		
		// Add N's when it's not the last contig in the scaffold...
		if (i < R.size()-1){			 
			int gapSize = (int)( R[i+1].mStart - R[i].mEnd);
			gapSize = (int)((double)gapSize * factor);
			int NumN = max(gapSize,OptMinN);
			//			NumN = (int)((double)NumN*factor);
			aStart=aStart+NumN;			 
			cout<<"Adding "<<NumN<<" Ns"<<flush;
			AddN(ScaffoldSeq,NumN);
			cout<<" totalsize="<<ScaffoldSeq.size()<<endl;
			AddZeros(ScaffoldQual,NumN);
		}				
	}
	if (pA != NULL) delete pA;

	
	// Don't fiddle with scaffolds smaller than some minimum...
	if (ScaffoldSeq.size() >= MinScaffoldSize){	
		SeqsOut.push_back(ScaffoldSeq);
		QualsOut.push_back(ScaffoldQual);
	}else{
		cerr<<"Omitting "<<ScaffoldSeq.mName<<" size="<<ScaffoldSeq.size()<<endl;
	}
}

//---------------------------------------------------
// * MeetsCriteria
//---------------------------------------------------
//
// link_contig_id
//
// linkid   varcar(30)
// contig   varcar(30)
// bacpair  number
// bacread  number 
// 
bool MeetsCriteria(BactigRecords &UnorderedScaffold){

	// KJD Accept everything now. 
	return(true); 

	String2IntMap BacPairMap;
	String2IntMap BacReadMap;

	// Grab a batch of the status from the DB (I go to the trouble
	// to grab a batch and stick it into a Map because it is too slow 
	// do do separate queries). 
	string bactig = UnorderedScaffold[0].mBactig;
	string query = "select * from link_contig_id where linkid='"+bactig+"'";
	cout<<"criteria query= "<<query<<endl;
	HGSCOracleStream q(DB,query);
	
	string linkid,contig;
	int bacpair,bacread;
	while(!q.eof()){
		q>>linkid>>contig>>bacpair>>bacread; 		
		string ID= linkid+"_"+contig;
		BacPairMap[ID] = bacpair;
		BacReadMap[ID] = bacread;
	}
	
	// See if each contig meets criteria...
	int failCount = 0;
	for(int c = 0;c< UnorderedScaffold.size();c++){
		string ID = UnorderedScaffold[c].mBactig+"_"+UnorderedScaffold[c].mContig;
		if (BacPairMap.Contains(ID)) bacpair = BacPairMap[ID];
		else {
			cerr<<ID<<" not in BacPairMap"<<endl;
			String2IntMap::iterator it;
			for(it = BacPairMap.begin();it != BacPairMap.end();it++){
				cerr<<"\t"<<(*it).first<<endl;
			}
		}

		if (BacReadMap.Contains(ID)) bacread = BacReadMap[ID];
		else cerr<<ID<<" not in BacReadMap"<<endl;

		if ((bacpair < 1) || (bacread < 2)) failCount++;
	}		
	
	if (failCount == UnorderedScaffold.size()){
		cerr<<"Omitting "<<UnorderedScaffold.mName<<" not criteria."<<endl;
		return(false);
	}else return(true);
}

//---------------------------------------------------
// * GetOrderedSubset
//---------------------------------------------------
//
void GetOrderedSubset(BactigRecords &R,BactigRecords &Ordered){
	Ordered.mName = R.mName;
	for(int i = 0;i< R.size();i++){
		if (R[i].mOrdered == '+'){
			Ordered.push_back(R[i]);
		}
	}
}

//---------------------------------------------------
// * GetUnorderedSubset
//---------------------------------------------------
//
void GetUnorderedSubset(BactigRecords &R,vector<BactigRecords> &Unordered){

	BactigRecords temp;	
	temp.mName = R.mName;
	for(int i = 0;i< R.size();i++){
		if (R[i].mOrdered == '-'){
			// When we reach a new unordered scaffold, push the previous
			// one onto the list, otherwise just keep adding contigs to the 
			// temporary record list. 
			if (R[i].mStart == 1){
				if (temp.size() > 0){
					Unordered.push_back(temp);
					temp.clear();
				}
				temp.push_back(R[i]);
			}else{
				temp.push_back(R[i]);
			}
		}
	}

	if (temp.size() > 0){
		Unordered.push_back(temp);
	}

}

//---------------------------------------------------
// * WriteScaffolds
//---------------------------------------------------
//
void WriteScaffolds(string OutName,vector<Scaffold> &TrueScaffolds)
{
	ofstream fout;
	fout.open(OutName.c_str());
	if (fout.fail()){
		cerr<<"Unable to open "<<OutName<<endl;
		return;
	}

	for(int i = 0;i< TrueScaffolds.size();i++){
		TrueScaffolds[i].Write(fout);
	}

	fout.close();
}

//---------------------------------------------------
// * RemoveX
//---------------------------------------------------
//
void RemoveX(SequenceVector &SeqsOut)
{
	for(int i = 0;i< SeqsOut.size();i++){
		for(int j = 0;j < SeqsOut[i].size();j++){
			if (toupper(SeqsOut[i][j]) == 'X'){
				SeqsOut[i][j] = 'N';
			}
		}
	}
}


//---------------------------------------------------
// * main
//---------------------------------------------------
//
int main(int argc, char** argv)
{

	OutRoot = argv[1];
	string SuperBactigID = argv[2];
	
	SequenceVector SeqsOut;
	SequenceQualityVector QualsOut;
	vector<Scaffold> TrueScaffolds;

	// Get the bactig records from the database and sort them...
	BactigRecords R;
	R.GetFromDB(DB,SuperBactigID);

	// Extract the subset of records that are part of the ordered scaffold...
	// and create a sequence for that. 
	BactigRecords Ordered;
	GetOrderedSubset(R,Ordered);

	int ScaffoldNumber = 0;
	if (Ordered.size() > 0){
		Scaffold OrderedScaffold;
		
		string snum = itoa(ScaffoldNumber);
		string name = Ordered.mName+"_Scaffold"+snum;
		string desc = "ORDERED";
		
		double factor = ComputeGapFactor(name,desc,Ordered);
		GetSequenceFromBactigRecords(name,desc,Ordered,SeqsOut,QualsOut,OrderedScaffold,factor);
		TrueScaffolds.push_back(OrderedScaffold);
	}
	
	// Now identify each of the unordered scaffolds...
	vector<BactigRecords> Unordered;
	GetUnorderedSubset(R,Unordered);
	
	// For each unordered scaffold that meets the criteria, get the sequence 
	// for that unordered scaffold...
	for(int i = 0;i< Unordered.size();i++){
		if (MeetsCriteria(Unordered[i])){
			ScaffoldNumber++;
			Scaffold UnorderedScaffold;
			
			string snum = itoa(ScaffoldNumber);
			string name = Unordered[i].mName+"_Scaffold"+snum;
			string desc = "UNORDERED "+Unordered[i][0].mBactig+"_"+Unordered[i][0].mScaffold;

			double factor = ComputeGapFactor(name,desc,Unordered[i]);
			GetSequenceFromBactigRecords(name,desc,Unordered[i],SeqsOut,QualsOut,UnorderedScaffold,factor);
			TrueScaffolds.push_back(UnorderedScaffold);
		}
	}
		 
	// Remove any Xs in the sequence (convert to N).
	RemoveX(SeqsOut);

	// Write out the sequence and quality....
	FASTAFile FA;
	string SeqOutName = OutRoot+"/"+SuperBactigID+"/"+SuperBactigID+".fa";
	FA.Write(SeqOutName,SeqsOut);
	string QualOutName = OutRoot+"/"+SuperBactigID+"/"+SuperBactigID+".qual";
	FA.Write(QualOutName,QualsOut);

	string ScaffoldOutName = OutRoot+"/"+SuperBactigID+"/"+SuperBactigID+".scaffold";
	WriteScaffolds(ScaffoldOutName,TrueScaffolds);

}





/*
SUPERBACTIG_SCAFFOLD
 Name                                      Null?    Type
 ----------------------------------------- -------- ----------------------------
 SUPERBACTIG                               NOT NULL VARCHAR2(10)
 BACTIG                                    NOT NULL VARCHAR2(10)
 SCAFFOLD                                  NOT NULL VARCHAR2(30)
 CONTIG                                    NOT NULL VARCHAR2(15)
 START_POS                                 NOT NULL NUMBER(10)
 END_POS                                   NOT NULL NUMBER(10)
 ORIENTATION                               NOT NULL CHAR(1)
 ORDERED                                   NOT NULL CHAR(1)
*/
