//==============================================================================
//	 
// atlas-binner.cc
//
// Binner for a whole-genome assembler.
// 1. Read in a graph, giving percentage overlap presumably based on sampling.
// 2. Construct an undirected graph where weights exceed cutoff.
// 3. Output connected components into separate bin files (singletons to bin 0).
//
//------------------------------------------------------------------------------
// $Id: //CBT-depot/CompBioToolsApps/tools/atlas-binner.cc#1 $
//------------------------------------------------------------------------------
// $Log: atlas-binner.cc,v $
// Revision 1.4  2003/02/12 09:33:27  havlak
// Now distinguish between left and right edges going from
// each node.  Edges into a node are not distinguished yet;
// up to 3*OptRptEdge inedges are allowed.
//
// Revision 1.3  2003/02/12 06:52:58  havlak
// More subtle repeats test; now checks at the repeat limit and cuts back
// to the point where there is a slope in the quality of overlaps.  That
// is, if the OptRptEdge value is 12, and the first edge and the 13th
// edge have the same weight, the node is treated as a repeat node.
// Otherwise, the edge list is resized to include only edges with weight
// > than the 13th node's.
//
// Revision 1.2  2003/02/12 05:10:50  havlak
// Modified options and score functions so that +w (estimated error-free
// span length) is default and -w means use UMD spans (confidence 0-100).
//
// Revision 1.1  2003/02/11 22:25:19  havlak
// Initial revision
//
// Revision 1.9  2002/11/08 17:28:18  havlak
// Parse and use frequency of seed for each overlap.
//
// Revision 1.8  2002/06/07 19:14:14  havlak
// Added +w for estimated-exact-span edge weighting.
//
// Revision 1.7  2001/10/12 20:06:04  havlak
// Modified computation of edge weight and support separate -i value.
//
// Revision 1.6  2001/09/24 19:06:02  havlak
// Added $Revision: #1 $ to help information to reveal version.
//
// Revision 1.5  2001/08/18 01:27:40  havlak
// Added -a option for alternate edge info -- in form of {span,score}
//
// Revision 1.4  2001/08/17 15:54:39  havlak
// Some old uninstalled changes plus beginnings of new revisions.
// 	-- use direct argument parsing
// 	-- different two-level binning (core bins plus satellites)
//
// Revision 1.3  2000/05/03 21:33:35  havlak
// Doesn't combine bins above minimum until last pass (thresh == ignore).
//
// Revision 1.2  2000/04/17 19:41:00  havlak
// Working with breadth-first search; no backlink check.
//
// Revision 1.1  2000/04/13 20:36:36  havlak
// Initial revision
//
// Revision 1.6  2000/04/10 22:15:32  havlak
// Working with wgs-overlap-precise
//
//==============================================================================
//.......1.........2.........3.........4.........5.........6.........7.........8
//

#include "LocalSTL.hh"
// #include "GetOption.hh"
// #include "Sequence.hh"
// #include "FASTAFile.hh"
// #include "FileUtils.hh"
#include <hash_map>
#include <strstream>
#include <string>
#include <cstdlib>

void myExit(int exitVal) {
  cout << flush;
  cerr << flush;
  exit(exitVal);
}

class Qinfo {
public:
  Qinfo(int nodeid, bool sameway) {
    encoded = nodeid;
    if (sameway) nodeid |= WAYMASK;
  }
  Qinfo() { encoded = -1; }
  bool isSameWay() { return (encoded & WAYMASK); }
  int node() { return (encoded & NODEMASK); }
private:
  static const int WAYMASK = 1 << ((8 * sizeof(int)) - 1);
  static const int NODEMASK = ~(WAYMASK);
  int encoded;
};

// Global Variables
const int PROCESSED = -1;	// placed in weight field to flag processed edges
string OptGraphFileName;
string OptOutPrefix;
string OptOutSuffix;
bool OptAllSinks;
bool OptSpanWeight;
string OptDebug;
int OptIgnore;
long OptNodes;
int OptBinMin;
int OptBinMax;
int OptBinDist;
int OptRptEdge;
int OptKillRpt;
int OptSinkRelative;
bool OptLRends;
bool OptWgsAll;
bool OptWalkThreshold;

struct Edge {
  int sink;	// name of the sink node (name of source is implicit)
  short weight;
  char samestrand;    

  Edge(int sk, short wt, char ss) { sink = sk; weight = wt; samestrand = ss; }

  Edge(const Edge& e) { sink = e.sink; weight = e.weight; samestrand = e.samestrand; }

  friend ostream& operator<<(ostream& os,const Edge &s);

  bool operator<(const Edge& e2) const {
    return weight > e2.weight;
  }
};

class EdgeVector: public vector<Edge> {
  // Make sure you can print out an edge list just like you can a string.
  friend ostream& operator<<(ostream& os,const EdgeVector &s){
    for(int i = 0;i<s.size();i++){
      os << " " << s[i];
    }
    return(os);
  }
};

const unsigned int INF_DIST = ~0;

// Colors:
const char BINWHITE = 0; // not binned, not queued
const char BINGRAY  = 1; // queued, temporarily binned
const char BINBLACK = 2; // binned 
const char BINGREEN = 3; // binned and printed
const char BINZERO  = 4; // binned and rejected to bin 0 (the junk bin for stuff too small)

struct Node {
  string name;	// name of the node
  short ins;    // number of in-edges, only computed at certain points
  char isWGS;   // short for a boolean, is this a true WGS read or not?
  char color;
  EdgeVector lefts;
  EdgeVector rights;

  // Node() { name = ""; ins = 0; }

  Node(const string& n) {
    long tempti;
    name = n; color = BINWHITE; ins = 0; 
    if (sscanf(n.c_string(), "gnl|ti|%ld", &tempti) 
	&& tempti >= 1000000000) {  // BES >= 1 billion
      isWGS = 0;
    }
    else {
      isWGS = 1;
    }
  }

  bool binned() { return (color != BINWHITE); }
  void unbin() { color = BINWHITE; }
  void bincolor(char c) { color = c; }

  friend ostream& operator<<(ostream& os,const Node &s){
    os << "\t" << s.name << "/" 
       << "w" << (s.isWGS ? "+" : "-") 
       << " bin" << (s.binned() ? "+" : "-") << 
      " L:" << s.lefts << " R:" << s.rights << endl;
    return(os);
  }
};

class NodeVector: public vector<Node> {
  // Make sure you can print out a graph just like you can a string.
  friend ostream& operator<<(ostream& os,const NodeVector &s){
    for(int i = 0;i<s.size();i++){
      os<<s[i];
    }
    return(os);
  }
};

NodeVector nodes;
char prefix[20] = "";

class NidVector: public vector<int> {
  friend ostream& operator <<(ostream& os, const NidVector &v) {
    for (int i = 0; i < v.size(); i++) {
      os << prefix << nodes[v[i]].name << endl;
    }
    return(os);
  }
};

// The friend << operator for Edges -- put here so they can print actual node names
ostream& operator<<(ostream& os,const Edge &s)
{
  os << nodes[s.sink].name << "=" << s.sink << (s.samestrand? '+' : '-') << s.weight;
  return(os);
}

// Hash stuff
struct eqstr
{
  bool operator()(string s1, string s2) const {
    return s1 == s2;
  }
};

struct hashstr {
  size_t operator()(string s) const {
    hash<const char *> H;
    return H(s.c_str());
  }
};

class StrHash: public hash_map<string, int, hashstr, eqstr> {
};
StrHash nMap;	// maps ReadName to index of node table

bool debugging(const char which[]) {
  if (OptDebug.find('+') != string::npos) return true;

  for (int i = 0; which[i]; i++) {
    if (OptDebug.find(which[i]) != string::npos) return true;
  }
  return false;
}

void PrintOptions() {
  cout << "Option values are:\n"<<
    "   -f {GraphFile}   ["<< OptGraphFileName << "] Name of graph file (empty => cin).\n"<<
    "   -p {BinPrefix}   ["<< OptOutPrefix <<"] Prefix for bin file names.\n"<<
    "   -s {BinSuffix}   ["<< OptOutSuffix <<"] Suffix for bin file names.\n"<<
    "   +A               ["<< (OptAllSinks? "+":"-") <<"] Boolean, put all sinks of core edges in one bin if true.\n"<<
    "   -i {Ignore}      ["<< OptIgnore <<"] Integer, minimum below which edge weight is ignored.\n" <<
    "   -r {RepeatEdge}  ["<< OptRptEdge <<"] Integer, max core edges before node treated as repeat.\n"<<
    "   -k {KillRepeat}  ["<< OptKillRpt <<"] Integer, kill overlap edges seeded by nmer with more than this many copies.\n"<<
    "   +S               ["<< (OptSinkRelative? "+":"-") <<"] Boolean, consider extensions relative to the sink (otherwise relative to origin).\n"<<
    "   +E               ["<< (OptLRends? "+":"-") <<"] Boolean, count left and right neighbors separtely against RepeatEdge count\n"<<
    "   -n {EstNodes}    ["<< OptNodes <<"] Integer, estimate max # nodes for efficiency's sake.\n" <<
    "   -m {MinBinSize}  ["<< OptBinMin <<"] Integer, minimum number of nodes per output bin.\n" <<
    "   -x {MaxBinSize}  ["<< OptBinMax <<"] Integer, maximum number of nodes per output bin.\n" <<
    "   -e {MaxBinDist}  ["<< OptBinDist <<"] Integer, maximum distance from start of bin in edges.\n" <<
    "   +w               ["<< (OptSpanWeight? "+":"-") << "] Boolean, use estimated-exact-span weighting.\n" <<
    "   -t               ["<< OptWalkThreshold << "] Integer, maximum size of search frontier before treated as repetitive region\n" <<
    "   +W               ["<< (OptWgsAll? "+":"-") << "] Boolean, treat all reads as WGS reads\n" <<
    "   -d {DebugString} ["<< OptDebug <<"] Each character indicates a debugging option to turn on;\n" <<
    "                            '+' indicates turn on all debugging.\n" <<
    "\n";
}

//---------------------------------------------------
// * PrintHelp
//---------------------------------------------------
//
void PrintHelp()
{
  cout<<"\n"<<
    "   atlas--binner	Reads in an asymmetric directed graph, with each\n" <<
    "			possible source node on a separate line, followed\n" <<
    "			by its whitespace-delimited outedges.  Each outedge\n" <<
    "			is represented as a pair, nodename whitespace weight,\n" <<
    "			where the weight is between 1 and 99.\n" <<
    "			Output is one file of read names (one read per line)\n" <<
    "			per bin.  Core reads of a bin occur in only one bin;\n" <<
    "			but peripheral reads may occur in multiple bins. \n" <<
    "			Bins are named prefixNNNNsuffix, where prefix0000suffix\n" <<
    "			is the default bin for reads not placed elsewhere\n" <<
    "			[Defaults are given in square braces.]\n" <<
    "			[$Revision: #1 $]\n";
  PrintOptions();
}

bool optOn(const string &inArg, char *option) {
  string optString(option);

  if (optString.find(inArg) == 0) return true; // matches starting at beginning
  else return false;
}

//---------------------------------------------------
// * SetupOptions
//---------------------------------------------------
//
int SetupOptions(int argc, char**argv)
{
  // Create parser object and define valid options
  // Options with a : require a parameter. 
  // GetOption Parser(argc,argv,"f:p:s:i:c:n:t:m:x:d:h");

  // Default values
  OptGraphFileName = "";	// 'f' Fail if not on command line
  OptOutPrefix = "bin";		// 'p' binNNNN.fon
  OptOutSuffix = ".fon";	// 's' (file of names)
  OptAllSinks = false;		// 'A'
  OptSpanWeight = true;		// 'w'
  OptIgnore    = 35;		// 'i' lesser weights treated as zero
  OptRptEdge   = 12;            // 'r' with more than this many edges assume node is repeat
  OptKillRpt   = 12;            // 'k' overlaps seeded by more abundant nmers ignored
  OptSinkRelative = true;       // 'S' treat extensions as relative to sink of each edge
  OptLRends = false;            // 'E' treat left and right extensions separately
  OptNodes = 1 << 18;		// 'n' 2^18 =~= 1 M / 4
  OptBinMin = 10;		// 'm' At least 10 reads per bin (smaller bins fused)
  OptBinMax = 99999;		// 'x' Limit on bin size
  OptBinDist = 10000;		// 'e' Limit on bin distance
  OptWalkThreshold = 12;
  OptWgsAll = false;		// 'W' treat all reads as WGS
  OptDebug = "";		// 'd'

  // Handle the options...
  int i;
  for (i = 1; i < argc; i++) {
    char prefix = argv[i][0];
    char theOption = argv[i][1];
    if ((prefix == '-') || (prefix == '+')) {
      bool sense = (prefix == '+');

      switch(theOption) {
      case 'f': OptGraphFileName = argv[++i]; break;
      case 'p': OptOutPrefix = argv[++i]; break;
      case 's': OptOutSuffix = argv[++i]; break;
      case 'A': OptAllSinks = sense; break;
      case 'w': OptSpanWeight = sense; break;
      case 'S': OptSinkRelative = sense; break;
      case 'E': OptLRends = sense; break;
      case 'i':
	OptIgnore = atoi(argv[++i]);
	if ((OptIgnore < 1) || (OptIgnore > 100)) {
	  cerr << "Ignore value " << OptIgnore << " should be in [1, 100]\n";
	  PrintHelp();
	  myExit(1);
	}
	break;
      case 'r':
	OptRptEdge = atoi(argv[++i]);
	if (OptRptEdge < 2) {
	  cerr << "Not useful to set edges repeat limit < 2\n";
	  PrintHelp();
	  myExit(1);
	}
	break;
      case 'k':
	OptKillRpt = atoi(argv[++i]);
	if (OptKillRpt < 3) {
	  cerr << "Not useful to set KillRpt nmer abundance limit < 3\n";
	  PrintHelp();
	  myExit(1);
	}
	break;
      case 'n':
	OptNodes = atoi(argv[++i]);
	if (OptNodes < 2) {
	  cerr << "Estimated number of nodes " << OptNodes << " should be > 1\n";
	  PrintHelp();
	  myExit(1);
	}
	break;
      case 't':
	break;
      case 'm':
	OptBinMin = atoi(argv[++i]);
	if (OptBinMin <= 1) {
	  cerr << "Minimum bin size " << OptBinMin << " should be > 1\n";
	  PrintHelp();
	  myExit(1);
	}
	if (OptBinMax <= OptBinMin) {
	  OptBinMax = OptBinMin*2;
	  cerr << "Maximum bin size reset to twice minimum: " << OptBinMax << endl;
	}
	break;
      case 'x':
	OptBinMax = atoi(argv[++i]);
	if (OptBinMax < 20) {
	  cerr << "Maximum bin size " << OptBinMax << " should be > 20.\n";
	  PrintHelp();
	  myExit(1);
	}
	if (OptBinMax <= OptBinMin) {
	  OptBinMin = OptBinMax/2;
	  cerr << "Minimum bin size reset to half maximum: " << OptBinMin << endl;
	}
	if (OptBinMin == 0) { OptBinMin = OptBinMax/2; }
	break;
      case 'e':
	OptBinDist = atoi(argv[++i]);
	if (OptBinDist < 1) {
	  cerr << "Minimum bin distance " << OptBinDist << " not allowed, using infinity.\n"
	       << endl;
	  OptBinDist = INF_DIST;
	}
	break;
      case 'd':
	OptDebug = argv[++i]; break;
      case 'h':
	PrintHelp(); myExit(0); 
	break;
      default: 
	cerr << "Unrecognized option: -" << theOption << "\n";
	PrintHelp(); 
	goto EndOptions;
      }
    }
  }
  EndOptions:
  if (debugging("o")) PrintOptions();
  return i;
}

int getNodeID(const string &name) {
  int id;

  if (nMap.count(name) == 0) {
    nodes.push_back(Node(name));
    nMap[name] = id = nodes.size() - 1;
  }
  else {
    id = nMap[name];
  }
  return id;
}

string getEdgeID(int sID, int dID) {
  char eID[50];
  sprintf(eID, "%d %d", sID, dID);
  return string(eID);
}

void readGraphFile(istream &graphIn) {
  string edgeLine;
  string sourceName;
  string destName;
  int sourceID;
  int destID;
  string edgeInfo;
  short weight;
  // double inWt;
  int eCnt = 0;
  int eUse = 0;

  while (getline(graphIn, edgeLine)) {

    istrstream edgeStream(edgeLine.c_str());
    
    edgeStream >> sourceName;

    sourceID = getNodeID(sourceName);

    // while ((edgeStream >> destName) && (edgeStream >> inWt)) {
    while ((edgeStream >> destName) && (edgeStream >> edgeInfo)) {
      destID = getNodeID(destName);
      // weight = min(100, int(inWt * 100));

      int span;
      int score;
      int lext;
      int rext;
      char dir;
      int freq;
      char junk[30];
      sscanf(edgeInfo.c_str(), "{%d,%d,%d,%d,%c,%d,%s", &span, &score, 
	     &lext, &rext, &dir, &freq, junk);

      if (OptSinkRelative) {
	if (dir == 'f') {
	  lext = -lext;
	  rext = -rext;
	}
	else { // reverse
	  int temp;
	  temp = lext;
	  lext = -rext;
	  rext = -temp;
	}
      }

      if (freq > OptKillRpt) continue; // seed too abundant

      if (OptSpanWeight) {
	// Want roughly the ratio of span to mismatches, adjusted
	//      * roughly, the expected largest mismatch-free span.
	//      * span-score is approximately twice the mismatch rate (mismatches are penalized
	//        twice, once by not getting their point and again losing a point for mismatch)
	//      * span/(1+ (span-score)/2) would give ratio if mismatches are evenly distributed
	//      * Use /3 instead of /2 to allow for less uniform distribution of mismatches
	//        (gives more advantage to longer matches with a few errors compared to short
	//        perfect)
	//      * ratio = span/(1+(span-score/3)) rewritten as below.
	weight = (int) (3.0*span/(3.0+span-score));
      }
      else {
	// Original scoring method
	// int delta = max(span - score, 0);
	// int adjscore = max(1, score - 32 - delta);
	// weight = (int) (-10 * log10((delta*delta + 0.5)/(adjscore*adjscore)));

	// Use UMD score as weight (in range 0..100).
	weight = score;
      }
      if (debugging("G")) {
	cout << sourceName << "/" << destName << "/" << edgeInfo 
	     << " " << weight << " " << (int)(score * 100.0/span) << endl;
      }

      if (destID == sourceID) {
//  	if (weight != 100) {
//  	  cerr << "Self edge on " << destName << " has weight " << 
//  	    weight << " less than 100!\n";
//  	}
	continue;
      }
      eCnt++;
      if (weight >= OptIgnore) {
	if (lext > 0 || ! OptLRends) {
	  (nodes[sourceID].lefts).push_back(Edge(destID, weight));
	}
	if (rext > 0 && OptLRends) {
	  (nodes[sourceID].rights).push_back(Edge(destID, weight));
	}
      	// nodes[destID].ins++;
	eUse++;
      }
    }
    if (debugging("G")) {
      if (! (eCnt % 10000)) cout << "." << flush;
    }
  }
  if (debugging("gG")) {
    cout << "\n\tNumber of nodes is: " << nodes.size() << "; ";
    cout << "\n\tEdges read is: " << eCnt << ", kept: " << eUse << endl;
    if (debugging("G")) {
      cout << nodes << endl;
    }
  }
}

ofstream binRpt;
ofstream binAll;

void print_bin(NidVector &bin, int num) {
  static char ofname[200];
  static int outBinSave = 1;	// external bin may group bins < BinMin
  int outBin;

  if (num == -1) {
    outBin = outBinSave++;
  }
  else {
    outBin = num;
  }

  binRpt << "Bin " << outBin << " with " << bin.size() << " nodes." << endl;

  if (! OptAllSinks) {
    sprintf(prefix, "%d ", outBin);
  }
  binAll << bin;

  if (debugging("p")) {
    cout << "\tBin contents: ";
    for (int i = 0; i < bin.size(); i++) {
      cout << nodes[bin[i]].name << " ";
      nodes[bin[i]].bincolor(BINGREEN);
    }
    cout << "\n";
  }
  bin.clear();
}

bool q_able(EdgeVector &ev) {
  int j;
  Edge t(0,0);
  int ne = ev.size();
  if (ne <= OptRptEdge) {
    return true;
  }
  sort(ev.begin(), ev.begin() + ne);

  // next stmt works because there are at least OptRptEdge+1 elements
  int beyond = ev[OptRptEdge].weight;
  for (j = OptRptEdge; j > 0; j--) { 
    if (ev[j-1].weight > beyond) break;
  }
  // This ensures we won't be sorting and double-checking every time.
  // Note that if there is no dropoff, we've made the list empty.
  ev.resize(j, t);
  // j > 0 implies a dropoff in weights, otherwise don't follow 
  // any of these edges
  return (j > 0);
}

void enqueue(const EdgeVector &ev) {
  int j;

  int ne = ev.size();

  for (j = 0; j < ne; j++) {
    int v = ev[j].sink;
    nodes[v].ins++;
  }
}

void binAllSinks() {
  NidVector allSinks;
  int nn = nodes.size();
  for (int n = 0; n < nn; n++) {
    if (q_able(nodes[n].lefts)) {
      enqueue(nodes[n].lefts);
    }
    if (q_able(nodes[n].rights)) {
      enqueue(nodes[n].rights);
    }
  }
  for (int i = 0; i < nn; i++) {
    if (nodes[i].ins > 0) {
      allSinks.push_back(i);
    }
  }
  print_bin(allSinks, -1);
}

bool process_q(deque &Q, NidVector &currBin) {
  int wgscount = 0;
  // If queue gets too big (in WGS), clear queue and return false to indicate abort
  // (should be pretty darn rare to stop just because there are no more links)
  //
  while (!(Q.empty())) {
    Qinfo qu = Q.front;
    int u = qu.node();
    Q.pop_front();	// remove u from Q
    
    if (qu.sameway) {
      neighbors = &(nodes[u].rights);
    }
    else {
      neighbors = &(nodes[u].lefts);
    }
    int ne = neighbors->size();
    
    for (int j = 0; j < ne; j++) {
      int v = (*neighbors)[j].sink;
      
      if (nodes[v].binned()) continue;
      if (nodes[v].isWGS) {
	wgscount++;
	if (wgscount > threshold) { // frontier too large, abort
	  nodes[u].unbin();
	  while (!(Q.empty())) {
	    Qinfo qx = Q.front();
	    x = qx.node();
	    nodes[x].unbin();	// make grays white again
	    Q.pop_front();
	  }
	  return false;
	}
      }
      nodes[v].bincolor(BINGRAY);

      sameway = ((qu.sameway && (*neighbors)[j].samestrand) ||
		 (!qu.sameway && !(*neighbors)[j].samestrand));
      
      if (sameway) { // go right
	if (q_able(nodes[v].rights)) {
	  Q.push_back(v, true);
	}
      }
      else { // go left
	if (q_able(nodes[v].lefts)) {
	  Q.push_back(v, false);
	}
      }
    } // finished with neighbors of u
    currBin.push_back(u);
    nodes[u].bincolor(BINBLACK);
  } // finished with queue
}

void bin_bfs() {

  int nn = nodes.size();
  deque <Qinfo> Q;
  NidVector currBin;
  NidVector tinyBin;
  EdgeVector *neighbors;
  bool sameway;

  binRpt << "Max, min bin sizes are: " << OptBinMax << ", " << OptBinMin << endl;
  binRpt << "Node count is: " << nodes.size() << endl;

  // Breadth-first search
  // white == not binned and not queued
  // gray  == binned and queued
  // black == binned and not queued

  for (int n = 0; n < nn; n++) {
    if (nodes[n].binned()) continue;
    nodes[n].bincolor(BINGRAY);		// start new bin here
    if (q_able(nodes[n].rights)) {
      Q.push_back(n, true);	// go right (downstream) from starting read
      // don't care for now if we stopped because of fork or running out of edges
      (void) process_q(Q, currBin);
    }
    if (q_able(nodes[n].lefts)) {
      Q.push_back(n, false);	// go left (upstream, sameway == false)
      // don't care for now if we stopped because of fork or running out of edges
      (void) process_q(Q, currBin);
    }
    // now Q is empty, all grays converted to black or white

    int B = currBin.size();
    if (B < OptBinMin) {
      // erase microbin if not at the last pass
      for (int i = 0; i < B; i++) {
	nodes[currBin[i]].bincolor(BINRED);
	tinyBin.push_back(i);
      }
      currBin.clear();
    }
    else {
      print_bin(currBin, -1);
      binRpt << "\tBin above minimum." << endl;
    }
  }
  print_bin(tinyBin, 0);	// Force bin number 0
  binRpt << "\tSubminimal bins." << endl;
}

extern "C" int unlink(const char *);

void rmOldBins() {
  char binName[200];
  int errCnt = 0;	// errors are negative, subtracting will increase this value
  int nremoved = 0;
  for (int i = 0; i < 9999; i++) {
    sprintf(binName, "%s%05d%s", OptOutPrefix.c_str(), i, OptOutSuffix.c_str());
    if (unlink(binName)) {
      errCnt++;
    } 
    else {
      nremoved++;
    }
    if (errCnt > 10) break;
  }
  cout << "Removed " << nremoved << " old bins";
  errCnt = nremoved = 0;
  for (int j = 10000; true; j++) {
    sprintf(binName, "%s%d%s", OptOutPrefix.c_str(), j, OptOutSuffix.c_str());
    if (unlink(binName)) {
      errCnt++;
    } 
    else {
      nremoved++;
    }
    if (errCnt > 20) break;
  }
  if (nremoved > 0) {
    cout << "(and " << nremoved << " more numbered over 9999)";
  }
  cout << "." << endl;
}

//------------------------------------------------------
// * main
//------------------------------------------------------
// 
int main(int argc, char**argv) {

  int firstArg = SetupOptions(argc,argv);

  // Get rid of any existing bin files -- be as restrictive as we can
  // in the removal pattern.
  rmOldBins();

  nodes.reserve(OptNodes);	// Reduce initial cost of growing.
  nMap.resize(OptNodes + OptNodes/4);

  // Keep the edge hash around only as long as necessary.
  cout << "Reading " << OptGraphFileName << "..." << flush;
  if (OptGraphFileName.length() > 0) {
    ifstream graphIn(OptGraphFileName.c_str());
    readGraphFile(graphIn);
    graphIn.close();
  }
  else { // cin
    if (firstArg >= argc) {
      readGraphFile(cin);
    }
    else { // list of graphs
      for (int i = firstArg; i < argc; i++) {
	ifstream graphIn(argv[i]);
	readGraphFile(graphIn);
	graphIn.close();
      }
    }
  }
  cout << "done." << endl;

  string binRptName = OptOutPrefix + OptOutSuffix + "log";
  string binAllName = OptOutPrefix + OptOutSuffix;
  binRpt.open(binRptName.c_str());
  binAll.open(binAllName.c_str());

  cout << "Making bins..." << flush;
  if (OptAllSinks) {
    binAllSinks();
  }
  else {
    bin_bfs();
  }
  binRpt.close();
  binAll.close();
  cout << "done." << endl;
}

